/**
 * This faceless Java applet makes it possible to connect a VO-compliant Web application
 * to a running local SAMP hub, allowing to broadcast VO-data to registered applications 
 * (e.g. Aladin, Splat, Topcat, ...). Notes that the applet must be signed, in particular
 * to enable the access to the SAMP configuration file ($HOME/.samp) on client's side. 
 * 
 * The communication between the JApplet and the Web browser is ensured by Javascript. The
 * JApplet has to be embedded in an <object> HTML tag in the client Web page (cf. 
 * http://www.w3schools.com/TAGS/tag_object.asp).
 *
 * @see http://vo.imcce.fr/webservices/wsc
 * @author J. Berthier (IMCCE/OBSPM) <berthier@imcce.fr>
 * @author R. Savalle (VOPDC/OBSPM) <renaud.savalle@obspm.fr>
 * @author A. Tregoubenko <http://arty.name>
 * @version 1.5, 2010-04-01
 * @copyright 2010, VO-Paris Data Centre <http://vo.obspm.fr/>
 * @license: CeCILL-C FREE SOFTWARE LICENSE <http://www.cecill.info> 
 */
package org.voparis;

import java.applet.Applet;
import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import netscape.javascript.JSObject;

import org.astrogrid.samp.Message;
import org.astrogrid.samp.Metadata;
import org.astrogrid.samp.SampUtils;
import org.astrogrid.samp.client.ClientProfile;
import org.astrogrid.samp.client.DefaultClientProfile;
import org.astrogrid.samp.client.HubConnection;
import org.astrogrid.samp.client.HubConnector;
import org.astrogrid.samp.client.SampException;
import org.astrogrid.samp.client.AbstractMessageHandler;

/**
 * VOParis Samp Connector class: communication between a Web browser and a VO software
 * using the IVOA Simple Application Messaging Protocol (SAMP).
 * @see <a href="http://vo.imcce.fr/webservices/samp">WebSampConnector Web site</a>
 * @author jberthier & rsavalle & arty
 * @version 1.5, 2010-04-01
 */
public class WebSampConnector extends Applet {

   /**
    * Default serial version ID of the type
    */
   private static final long serialVersionUID = 7022485897611517032L;

   /**
    * Defines Profile-specific aspects of the SAMP client implementation
    */
   private ClientProfile profile;

   /**
    * Manages a client's connection to SAMP hub
    */
   private HubConnector connector;

   /**
    * Represents a registered client's connection to a running hub
    */
   private HubConnection connection;

   /**
    * Javascript object used to communicate with the client
    */
   private static JSObject clientWSC;

   /**
    * Init method called by the browser to register the client and to connect to the hub. 
    * The method ends with the call of the client's callback function named 'onlineHandler'.
    */
   @Override
   public void init() {

      super.init();

      try {
         // Gets client's window and proxy object 
         JSObject clientWindow = JSObject.getWindow(this);
         WebSampConnector.clientWSC = (JSObject) clientWindow.getMember("WebSampConnector");
         // Client default profile
         this.profile = DefaultClientProfile.getProfile();
         // Connection object
         this.connector = new HubConnector(this.profile);
         // Declares the metadata of the VOPSampConnector
         this.connector.declareMetadata(getConnectorMetadata());
         // Install a handler for SAMP msg coord.pointAt.sky
         this.connector.addMessageHandler(new pointAtSkyHandler());
         // Install a handler for the SAMP msg table.highlight.row
         this.connector.addMessageHandler(new highlightRowHandler());
         // Install a handler for SAMP msg table.select.rowList
         this.connector.addMessageHandler(new selectRowListHandler());
         // Install a handler for SAMP msg spectrum.load.ssa-generic
         this.connector.addMessageHandler(new spectrumLoadSsaHandler());
         // Declares the MType subscriptions for this client
         this.connector.declareSubscriptions(this.connector.computeSubscriptions());
         // Hub connection representing configured registration with a hub if it is running 
         this.connection = this.connector.getConnection();
         // Get the status of the connection
         String sConn = (this.connector.isConnected()) ? "connected" : "disconnected";
         // Call the client's callback function
         WebSampConnector.clientWSC.call("onlineHandler", new Object[] {sConn});
      } catch (SampException se) {
         PrintErrorMsg(se);
      }

   }

   /**
    * Destructor method to explicitly disconnect the client from the hub
    */
   @Override
   public void destroy() {
      
      this.connector.setActive(false);
      
   }   

   /**
    * This method unregisters the client, terminates the connection and performs
    * any associated required cleanup.
    * @return true if no error occurs, if not false
    * @throws PrivilegedActionException 
    */
   public boolean disconnect() throws PrivilegedActionException {
      final HubConnector connector = this.connector;
      Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
         public Object run() {
            connector.setActive(false);
            return true;
         }
      }));
      Boolean result = (Boolean) obj;
      return result.booleanValue();
   }

   /**
    * This method allows the client to know if he is connected or not to a Samp hub
    * @return true if connected, if not false
    * @throws PrivilegedActionException
    */
   public boolean isConnected() throws PrivilegedActionException {
      final HubConnector connector = this.connector;
      Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
         public Object run() {
            return connector.isConnected();
         }
      }));
      Boolean result = (Boolean) obj;
      return result.booleanValue();
   }

   /**
    * This method retrieves the list of client public IDs for those clients currently registered.
    * @return array containing the list of registered clients and their metadata as a WebSampConnectorClient object
    * @throws PrivilegedActionException
    */
   public Object[] getRegisteredClients() throws PrivilegedActionException {
      final HubConnection connection = this.connection;
      Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
         public Object run() {
            Object[] res;
            try {
               String[] clients = connection.getRegisteredClients();
               res = new Object[clients.length];
               for (int i=0; i<clients.length; i++) {
                  Metadata meta = new Metadata(connection.getMetadata(clients[i]));
                  res[i] = new WebSampConnectorClient(clients[i], meta);
               }
            } catch (SampException e) {
               res = new Object[] {"Error:" + e.getStackTrace()};
            }
            return res;
         }
      }));
      Object[] result = (Object[]) obj;
      return result;
   }

   /**
    * This method retrieves the list of clients which subscribed a given MType
    * @param mtype the given MType (e.g. table.load.votable)
    * @return array containing the list of clients which subscribed the given MType, and their metadata as a WebSampConnectorClient object
    * @throws PrivilegedActionException
    */
   public Object[] getSubscribedClients(final String mtype) throws PrivilegedActionException {
      final HubConnection connection = this.connection;
      Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
         public Object run() {
            Object[] res;
            try {
               Map<String,String> map = new HashMap<String,String>();
               map = connection.getSubscribedClients(mtype);
               int mapsize = map.size();
               res = new Object[mapsize];
               int i = 0;
               for (String mapKey : map.keySet()) {
                  Metadata meta = new Metadata(connection.getMetadata(mapKey));
                  res[i] = new WebSampConnectorClient(mapKey, meta);
                  i++;
               }
            } catch (SampException e) {
               res = new String[] {"Error:" + e.getStackTrace()};
            }
            return res;
         }
      }));
      Object[] result = (Object[]) obj;
      return result;
   }
   
   /**
    * This method allows the client to broadcast data to VO-compliant applications
    * @see <a href="http://www.ivoa.net/cgi-bin/twiki/bin/view/IVOA/SampMTypes">Samp MTypes</a>
    * @param mType the MType of the resource (e.g. table.load.votable)
    * @param id the Id of the resource
    * @param name the name of the resource
    * @param url the URL where is located the resource
    * @param keyName key associated to the resource id (param id) which may be used to refer to the loaded resource in subsequent messages
    * @return true if no error occurs, if not false (if an exception is rises then a message is printed)
    * @throws PrivilegedActionException 
    */
   public boolean sendMsg(final String mType, final String id, final String name, final String url, final String keyName) throws PrivilegedActionException {
      final HubConnection connection = this.connection;
      Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
         public Object run() {
            // Prepares the SAMP msg
            String idKey = null;
            if (keyName.isEmpty()) {
               if (mType.equals("table.load.votable")) {
                  idKey = "table-id";
               } else if (mType.equals("table.load.fits")) {
                  idKey = "table-id";
               } else if (mType.equals("image.load.fits")) {
                  idKey = "image-id";
               } else if (mType.equals("spectrum.load.ssa-generic")) {
                  idKey = "spectrum-id";
               }
            } else {
               idKey = keyName;
            }
            // Calls JSAMP
            try {
               // Encodes the SAMP message
               Message msg = new Message(mType);
               msg.addParam("url", url);
               msg.addParam("name", name);
               if (idKey != null) {
                  msg.addParam(idKey, id);
               }
               // Sends a message to all subscribed clients without wanting a response
               connection.notifyAll(msg);
               return true;
            } catch (SampException se) {
               PrintErrorMsg(se);
               return false;
            }
         }
      }));
      Boolean result = (Boolean) obj;
      return result.booleanValue();
   }

   // --------------------------------------------------------------------------------------------------
   // --- The following methods implement specific MTypes for interaction with VO-compliant applications
   // --------------------------------------------------------------------------------------------------

   /**
    * This method allows the client to point a given celestial coordinates
    * @param ra the right ascension to point (in degrees)
    * @param dec the declination to point (in degrees)
    * @return true if no error occurs, if not false
    * @throws PrivilegedActionException 
    */
   public boolean pointAtSky(final double ra, final double dec) throws PrivilegedActionException {
      final HubConnection connection = this.connection;
      Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
         public Object run() {
            try {
               // Prepares the SAMP msg
               String mtype = "coord.pointAt.sky";
               // Encodes the SAMP message
               Message msg = new Message(mtype);
               msg.addParam("ra", SampUtils.encodeFloat(ra));
               msg.addParam("dec", SampUtils.encodeFloat(dec));
               // Sends a message to all subscribed clients without wanting a response
               connection.notifyAll(msg);
               return true;
            } catch (SampException se) {
               PrintErrorMsg(se);
               return false;
            }
         }
      }));
      Boolean result = (Boolean) obj;
      return result.booleanValue();
   }

   /**
    * This method allows the client to highlight a given row of a table
    * @param tableId the Id of the resource (table)
    * @param url the URL where is located the resource
    * @param row the number of the row to highlight (from 0 to n-1)
    * @return true if no error occurs, if not false
    * @throws PrivilegedActionException 
    */
   public boolean tableHighlightRow(final String tableId, final String url, final int row) throws PrivilegedActionException {
      final HubConnection connection = this.connection;
      Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
         public Object run() {
            try {
               // Prepares the SAMP msg
               String mtype = "table.highlight.row";
               // Encodes the SAMP message
               Message msg = new Message(mtype);
               msg.addParam("table-id", tableId);
               msg.addParam("url", url);
               msg.addParam("row", SampUtils.encodeInt(row));
               // Sends a message to all subscribed clients without wanting a response
               connection.notifyAll(msg);
               return true;
            } catch (SampException se) {
               PrintErrorMsg(se);
               return false;
            }
         }
      }));
      Boolean result = (Boolean) obj;
      return result.booleanValue();
   }

   /**
    * This method allows the client to highlight a set of rows of the table
    * @param tableId the Id of the resource (table)
    * @param url the URL where is located the resource
    * @param rowList an array of integer providing the list of the row numbers to highlight (from 0 to n-1)
    * @return true if no error occurs, if not false
    * @throws PrivilegedActionException 
    */
   public boolean tableSelectRowList(final String tableId, final String url, final int[] rowList) throws PrivilegedActionException {
      final HubConnection connection = this.connection;
      Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
         public Object run() {
            try {
               // Prepares the SAMP msg
               String mtype = "table.select.rowList";
               // Encodes the SAMP message
               Message msg = new Message(mtype);
               msg.addParam("table-id", tableId);
               msg.addParam("url", url);
               ArrayList<String> rList = new ArrayList<String>();
               for (int i=0; i<rowList.length; i++) {
                  rList.add(SampUtils.encodeInt(rowList[i]));
               }
               msg.addParam("row-list", rList);
               // Sends a message to all subscribed clients without wanting a response
               connection.notifyAll(msg);
               return true;
            } catch (SampException se) {
               PrintErrorMsg(se);
               return false;
            }
         }
      }));
      Boolean result = (Boolean) obj;
      return result.booleanValue();
   }

   /**
    * This method allows the client to send a script to be executed by Aladin
    * @param script a string containing the Aladin script to be executed
    * @return true if no error occurs, if not false
    * @throws PrivilegedActionException 
    */
   public boolean sendAladinScript(final String script) throws PrivilegedActionException {
      final HubConnection connection = this.connection;
      Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
         public Object run() {
            try {
               // Prepares the SAMP msg
               String mtype = "script.aladin.send";
               // Encodes the SAMP message
               Message msg = new Message(mtype);
               msg.addParam("script", script);
               // Sends a message to all subscribed clients without wanting a response
               connection.notifyAll(msg);
               return true;
            } catch (SampException se) {
               PrintErrorMsg(se);
               return false;
            }
         }
      }));
      Boolean result = (Boolean) obj;
      return result.booleanValue();
   }

   // --------------------------------------------------------------------------------
   // --- The following methods process the call from applications for specific MTypes
   // --------------------------------------------------------------------------------

   /**
    * This class subscribes the coord.pointAt.sky message handler of SAMP, 
    * and implements the processCall() method of the abstract massage handler class.
    * The data received from the application are forwarded to the client as an array. 
    */
   public static class pointAtSkyHandler extends AbstractMessageHandler {

      pointAtSkyHandler() {
         super("coord.pointAt.sky");
      }

      public Map processCall(final HubConnection connection, final String senderId, final Message msg) throws Exception {
         Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
            public Object run() {
               // Retrieves parameters from call (celestial coord. in degrees)
               String ra = (String) msg.getParam("ra");
               String de = (String) msg.getParam("dec");
               double fra = SampUtils.decodeFloat(ra);
               double fde = SampUtils.decodeFloat(de);
               // Sends to the client the RA,DEC coordinates to point at
               if (WebSampConnector.clientWSC != null) {
                  Object[] args = new Object[] {""+fra, ""+fde};
                  return (JSObject) WebSampConnector.clientWSC.call("pointAtSkyHandler", args);
               } else {
                  return null;
               }
            }
         }));
         return null;
      }
   }

   /**
    * This class subscribes the table highlight row message handler of SAMP, 
    * and implements the processCall() method of the abstract massage handler class. 
    * The data received from the application are forwarded to the client as an array.
    */
   public static class highlightRowHandler extends AbstractMessageHandler {

      highlightRowHandler() {
         super("table.highlight.row");
      }

      public Map processCall(final HubConnection connection, final String senderId, final Message msg) throws Exception {
          Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
             public Object run() {
                // Retrieves parameters from call (celestial coord. in degrees)
                String tId = (String) msg.getParam("table-id");
                String url = (String) msg.getParam("url");
                String row = (String) msg.getParam("row");
                // Sends to the client the row to highlight
                if (WebSampConnector.clientWSC != null) {
                   Object[] args = new Object[] {tId, url, row};
                   return (JSObject) WebSampConnector.clientWSC.call("highlightRowHandler", args);
                } else {
                   return null;
                }
             }
          }));
          return null;
       }
   }

   /**
    * This class subscribes the select row list message handler of SAMP, 
    * and implements the processCall() method of the abstract massage handler class. 
    * The data received from the application are forwarded to the client as an array.
    */
   public static class selectRowListHandler extends AbstractMessageHandler {

      selectRowListHandler() {
         super("table.select.rowList");
      }

      public Map processCall(final HubConnection connection, final String senderId, final Message msg) throws Exception {
          Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
             public Object run() {
                // Retrieves parameters from call (celestial coord. in degrees)
                String tId = (String) msg.getParam("table-id");
                String url = (String) msg.getParam("url");
                List rows = (List) msg.getParam("row-list");
                // Sends to the client the list of rows
                if (WebSampConnector.clientWSC != null) {
                   String[] stringRows = new String[rows.size()];
                   int i = 0;
                   for (Iterator it = rows.iterator(); it.hasNext();) {
                       stringRows[i] = (String) it.next();
                       i++;
                    }
                   Object[] args = new Object[] {tId, url, stringRows};
                   return (JSObject) WebSampConnector.clientWSC.call("selectRowListHandler", args);
                } else {
                   return null;
                }
             }
          }));
          return null;
       }
   }

   /**
    * This class subscribes the spectrum.load.ssa-generic message handler of SAMP, 
    * and implements the processCall() method of the abstract massage handler class. 
    * The data received from the application are forwarded to the client as an array.
    */
   public static class spectrumLoadSsaHandler extends AbstractMessageHandler {

      spectrumLoadSsaHandler() {
         super("spectrum.load.ssa-generic");
      }

      public Map processCall(final HubConnection connection, final String senderId, final Message msg) throws Exception {
          Object obj = (AccessController.doPrivileged(new PrivilegedExceptionAction<Object>() {
             public Object run() {
                // Retrieves parameters from call
                String url = (String) msg.getParam("url");
                String sId = (String) msg.getParam("spectrum-id");
                String name = (String) msg.getParam("name");
                // Sends to the client the list of rows
                if (WebSampConnector.clientWSC != null) {
                   Object[] args = new Object[] {url, sId, name};
                   return (JSObject) WebSampConnector.clientWSC.call("spectrumLoadSsaHandler", args);
                } else {
                   return null;
                }
             }
          }));
          return null;
       }
   }

   // --------------------------------------------------------------
   // --- The following methods are private methods for internal use
   // --------------------------------------------------------------

   /**
    * This method defines the metadata map of the WebSampConnector
    * @return the map of the WebSampConnector metadata
    */
   private Metadata getConnectorMetadata() {
      Metadata map = new Metadata();
      // Sets the value for the application's name
      map.setName("WebSampConnector");
      // Sets a short description of the application.
      map.setDescriptionText("Web browser Samp Connector");
      // Sets an HTML description of the application
      map.setDescriptionHtml("Connects your Web browser to a runnig SAMP hub to broadcast data to VO-compliant applications");
      // Sets a URL for a documentation web page.
      map.setDocumentationUrl("http://vo.imcce.fr/webservices/samp/");
      // Sets a URL for a gif, png or jpeg icon identifying the application
      map.setIconUrl("http://vo.imcce.fr/webservices/share/logo/logo_VOParis.png");
      // Return the map
      return map;
   }

   /**
    * This class defines the metadata of a client as seen from the Samp hub.
    */
   public class WebSampConnectorClient {
      /**
       * Public identifier for this client
       */
      public String id;
      /**
       * Application's name
       */
      public String name;
      /**
       * Short description of the application
       */
      public String descriptionText;
      /**
       * HTML description of the application
       */
      public String descriptionHtml;
      /**
       * URL for a gif, png or jpeg icon identifying the application
       */
      public String iconUrl;
      /**
       * URL for a documentation web page
       */
      public String documentationUrl;

      /**
       * Class constructor
       * @param client public Id of the client
       * @param meta metadata of the client
       */
      public WebSampConnectorClient(String client, Metadata meta) {
         this.id = client;
         this.name = meta.getName();
         this.descriptionText = meta.getDescriptionText();
         this.descriptionHtml = meta.getDescriptionHtml();
         java.net.URL icon = meta.getIconUrl();
         if (icon != null) {
            this.iconUrl = icon.toString();
         }
         java.net.URL docs = meta.getDocumentationUrl();
         if (docs != null) {
            this.documentationUrl = meta.getDocumentationUrl().toString();
         }
      }
   }

   /**
    * This method prints out an exception message
    * @param e the exception
    */
   private void PrintErrorMsg(Exception e) {
      System.out.println("Error: WebSampConnector:");
      if (e.getMessage() != null) {
         System.out.println("  Message: " + e.getMessage());
      }
   }

}
